<?php

namespace HighWayPro\App\HighWayPro\DestinationComponents\BuiltIn\Conditions;

use HighWayPro\App\HighWayPro\DestinationComponents\BuiltIn\Conditions\Errors\DisposableConditionErrors;
use HighWayPro\App\HighWayPro\DestinationComponents\BuiltIn\Conditions\Errors\HourRangeConditionErrors;
use HighWayPro\App\HighWayPro\DestinationComponents\DestinationConditionComponent;
use HighWayPro\Original\Collections\Collection;
use HighWayPro\Original\Environment\Env;
use HighwayPro\App\HighwayPro\System\Date;
use HighwayPro\Original\Collections\Mapper\Types;

Class HourRangeCondition extends DestinationConditionComponent
{
    const name = 'highwaypro.HourRangeCondition';

    /**
     * Optional, used in testing environments
     * @var callable
     */
    public $currentTimeFunction;

    public static function title()
    {
        return __('Hour Range', Env::textDomain());
    }

    public static function shortDescription()
    {
        return __('Limit the time of the day the destination can be accessed.', Env::textDomain());
    }

    public static function description()
    {
        return __(
            "Limit access to the destination based on the time of the day, in hours. For example, you can make the destination accessible only from 2 to 6 PM, everyday.\n".
            "The destination will be accesible before the end hour. For example, if you select from 2 to 6 pm, the destination will be accessible from 2:01 to 5:59 pm.\n".
            "You can select one hour only (for example, 2 pm), by selecting the same start and end hours.\n". 
            "You can also select cross-day ranges. For example, from 8 pm to 5 am.",
            Env::textDomain()
        );
    }
    
    protected function parametersMap()
    {
        return [
            'dayHours' => [
                'start' => Types::INTEGER()->withDefault(0)->allowed(range(0, 23)),
                'end'   => Types::INTEGER()->withDefault(0)->allowed(range(0, 23))
            ]
        ];
    }

    public function hasPassed()
    {
        (object) $currentDate = Date::createFromSettings($this->currentTimeFunction);

        (object) $allowedHours = Date::getDayHoursRange(
            $this->parameters->dayHours->start, $this->parameters->dayHours->end
        );

        if ($allowedHours->haveExactly(1)) {
            return $allowedHours->contain($currentDate->getHour());
        } 

        return $allowedHours->last() > 0 && $allowedHours->withoutLast()->contain($currentDate->getHour());
    }

    public function validateParameters()
    {
        if ($this->parameters->mapFieldsNotFoundInSource->contain('dayHours')) {
            return HourRangeConditionErrors::get('no_hour_start');
        } elseif (!$this->parameters->dataFound->dayHours->mapFieldsFoundInSource->contain('start')) {
            return HourRangeConditionErrors::get('no_hour_start');
        } elseif (!$this->parameters->dataFound->dayHours->mapFieldsFoundInSource->contain('end')) {
            return HourRangeConditionErrors::get('no_hour_end');
        }

        return true;
    }
}