<?php

namespace HighWayPro\App\HighwayPro\Time;

use Exception;
use HighWayPro\Original\Characters\StringManager;
use HighWayPro\Original\Collections\Collection;
use HighwayPro\App\HighwayPro\System\Date;
use InvalidArgumentException;

Class PeriodOfTime
{
    public static $validUnits = [ 'auto', 'hours', 'days', 'weeks', 'months'];

    protected $range;
    protected $forcedUnit; // days, weeks, months 

    /**
     * $today MUST be an object with the day of the current day (aka TODAY)
     *
     * Presets are case-insensitive.
     * 
     * if 'customrange' is set, then $customRange must be an array of: 
     * [
     *     'from' => (string) Date::FORMAT_NO_TIMESTAMP,
     *     'to' => (string) Date::FORMAT_NO_TIMESTAMP
     * ]
     */
    public static function createFromPreset($preset, Date $today, $forcedUnit = null, array $customRange = [])
    {
        (string) $from = '';
        (string) $to = '';

        switch (strtolower($preset)) {
            case PeriodOfTimePresets::TODAY:
                $from = $today->copy()->startOfDay()->format(Date::FORMAT);
                $to = $today->copy()->endOfDay()->format(Date::FORMAT);
                break;
            case PeriodOfTimePresets::LAST_DAY:
                $from = $today->copy()->subDay()->startOfDay()->format(Date::FORMAT);
                $to = $today->copy()->subDay()->endOfDay()->format(Date::FORMAT);
                break;
            case PeriodOfTimePresets::TWO_DAYS_AGO: 
                // NOTE: TWO DAYS AGO AND IN THE LAST TWO DAYS (LIKE IN_THE_LAST_SEVEN_DAYS BELOW)
                // ARE COMPLETELY DIFFERENT THINGS. ONE REFERS TO A SINGLE DAY AND THE OTHER TO 
                // A PERIOD FROM THE N DAYS IN THE PAST TO THE CURRENT DATE
                $from = $today->copy()->subDays(2)->startOfDay()->format(Date::FORMAT);
                $to = $today->copy()->subDays(2)->endOfDay()->format(Date::FORMAT);;
                break;
            case PeriodOfTimePresets::IN_THE_LAST_SEVEN_DAYS:
                $from = $today->copy()->subDays(7)->startOfDay()->format(Date::FORMAT);
                $to = $today->copy()->startOfDay()->format(Date::FORMAT);
                break;
            case PeriodOfTimePresets::IN_THE_LAST_MONTH:
                $from = $today->copy()->subMonthNoOverflow()->startOfDay()->format(Date::FORMAT);
                $to = $today->copy()->endOfDay()->format(Date::FORMAT);
                break;
            case PeriodOfTimePresets::IN_THE_LAST_YEAR:
                $from = $today->copy()->subYearNoOverflow()->startOfDay()->format(Date::FORMAT);
                $to = $today->copy()->endOfDay()->format(Date::FORMAT);
                break;
            case PeriodOfTimePresets::THIS_MONTH:
                $from = $today->copy()->startOfMonth()->startOfDay()->format(Date::FORMAT);
                $to = $today->copy()->endOfDay()->format(Date::FORMAT);
                break;
            case PeriodOfTimePresets::THIS_YEAR:
                $from = $today->copy()->startOfYear()->startOfDay()->format(Date::FORMAT);
                $to = $today->copy()->endOfDay()->format(Date::FORMAT);
                break;
            case PeriodOfTimePresets::SAME_PERIOD_OF_LAST_MONTH:
                $from = $today->copy()
                              ->subMonthNoOverflow()
                              ->startOfMonth()
                              ->startOfDay()
                              ->format(Date::FORMAT);
                $to = $today->copy()
                            ->subMonthNoOverflow()
                            ->endOfDay()
                            ->format(Date::FORMAT);
                break;
            case PeriodOfTimePresets::SAME_PERIOD_OF_LAST_YEAR:
                $from = $today->copy()
                              ->subYearNoOverflow()
                              ->startOfYear()
                              ->startOfDay()
                              ->format(Date::FORMAT);
                $to = $today->copy()
                            ->subYearNoOverflow()
                            ->endOfDay()
                            ->format(Date::FORMAT);
                break;
            case PeriodOfTimePresets::CUSTOM_RANGE:
                $from = $customRange['from'];
                $to = $customRange['to'];
                break;
            default:
                throw new InvalidArgumentException("Invalid PeriodOfTime preset: {$preset}.");
                break;
        }

        return new static($from, $to, $forcedUnit);
    }

    public function __construct($from, $to, $forcedUnit = null)
    {
        $this->range = new Collection([]);

        $this->range->set('from', Date::createFromString($from));
        $this->range->set('to', Date::createFromString($to));

        $this->forcedUnit = $forcedUnit;

        $this->validateArgments($forcedUnit);
    }

    public function getLength($forcedUnit = null)
    {
        //so, what were going to do here is: If range is hours, then get the total number of hours in selected range (eg, if one day, then 24, if 3 days, then 72 hours, etc)
        // if range is days, then get the total number of days in the selected range.
            //PLEASE NOT THAT MONTHS have different ammounts. Four in fact:
                // 28, 29, 30, 31 (28 and 29 are from febr.) so a range from feb 1 to, say, mar 30 **WON'T** BE 60 days.
        (object) $from = $this->range->get('from');
        (object) $to = $this->range->get('to');
        (integer) $theCurrentDay = 1;

        switch ($forcedUnit?: $this->getUnit()) {
            case 'hours':
                return 24;
                break;
            case 'days':
                return $from->diffinDays($to) + $theCurrentDay;
            case 'weeks':
                return $from->diffInWeeks($to) + $theCurrentDay;
            case 'months':
                return $from->diffInMonths($to) + $theCurrentDay;
                break;
        }
    }

    public function unitIs($unit)
    {
        return $this->getUnit() === $unit;   
    }
    
    public function getUnit()
    {
        (object) $from = $this->range->get('from')->copy()->startOfDay();
        (object) $to = $this->range->get('to')->copy()->startOfDay();

        if ($from->isSameDay($to)) {
            return 'hours';
        } elseif ($this->hasForcedAUnit()) {
            return $this->forcedUnit;
        } elseif ($to->lessThanOrEqualTo($from->copy()->addMonthsNoOverflow(1))) {
            // less than a month or exactly month
            return 'days';
        } elseif  ($to->lessThanOrEqualTo($from->copy()->addMonthsNoOverflow(2))) {
            // between 1 and two months
            return 'weeks';
        } else {
            // more than two months...
            return 'months';
        }
    }

    /**
     * Please note, getStartDate and getEndDate dates are modified to the 
     * first and last time of the days
     * this has been designed  so that we get all the rows regardless of the time of the day
     */
    public function getStartDate()
    {
        return $this->range->get('from')->copy()->startOfDay();
    }
    
    public function getEndDate()
    {
        return $this->range->get('to')->copy()->endOfDay();
    }

    public function getStartDateFormattedWithTimestamp()
    {
        return $this->getStartDate()->format(Date::FORMAT);   
    }
    
    public function getEndDateFormattedWithTimestamp()
    {
        return $this->getEndDate()->format(Date::FORMAT);   
    }

    public function getStartDateFormatted()
    {
        return $this->getStartDate()->format(Date::FORMAT_NO_TIMESTAMP);   
    }
    
    public function getEndDateFormatted()
    {
        return $this->getEndDate()->format(Date::FORMAT_NO_TIMESTAMP);   
    }

    public function getDateRanges()
    {
        (object) $from = $this->range->get('from');
        (object) $to = $this->range->get('to');

        switch ($this->getUnit()) {
            case 'hours':
                return $this->getHoursFromDateRange();
                break;
            case 'days':
                return $this->getDaysFromDateRange();
                break;
            case 'weeks':
                return $this->getWeeksFromDateRange();
                break;
            case 'months':
                return $this->getMonthsFromDateRange();
                break;
            
            default:
                throw new Exception("Unit not supported as date range: {$this->getUnit()}");
                break;
        }
    }
    
    public function getDaysFromDateRange()
    {
        (integer) $currentIndex = -1;
        (object) $from = $this->range->get('from');

        return Collection::range(1, $this->getLength('days'))->map(function() use (&$currentIndex, $from) {
            $currentIndex++;
            return $from->copy()->addDays($currentIndex);
        });  
    }

    /**
     * Currently, only returns the (24) hours from the beggining to the end of the day
     */
    public function getHoursFromDateRange()
    {
        (integer) $currentIndex = -1;
        (object) $from = $this->range->get('from')->startOfDay();
        (object) $to = $this->range->get('to')->endOfDay();

        return Collection::range(1, $this->getLength())->map(function() use (&$currentIndex, $from) {
            $currentIndex++;
            return $from->copy()->addHours($currentIndex);
        });  
    }

    public function getWeeksFromDateRange()
    {
        (integer) $currentIndex = -1;
        (object) $from = $this->getStartDate();
        (object) $to = $this->getEndDate();

        return Collection::range(1, $from->diffInWeeks($to) + 1)->map(function() use (&$currentIndex, $from, $to) {
            $currentIndex++;
            (object) $currentFrom = $from->copy();
            (object) $sevenDaysLater = $from->addDays(6);
            // we may not need full weeks, for example, from oct 02 to nov 05, there's an incomplete week, 
            // we want to return partial weeks if needed, not always 7-day weeks.
            (object) $nextWeek = $sevenDaysLater->min($to);

            $week = new Collection([
                'from' => $currentFrom->copy()->startOfDay(), // range should start from the very firts second of the day
                'to' => $nextWeek->copy()->endOfDay() // and end on the last second of the max day (23:59:59)
            ]);

            $from->addDays(1);

            return $week;
        });  
    }

    public function getMonthsFromDateRange()
    {
        (integer) $currentIndex = -1;
        (object) $from = $this->getStartDate();
        (object) $to = $this->getEndDate();

        return Collection::range(1, $this->getTheTotalNumberOfMonthsInRange())->map(function() use (&$currentIndex, $from, $to) {
            $currentIndex++;

            (object) $currentDate = $from->copy()->addMonthsNoOverflow($currentIndex);
            (object) $endOfCurrentMonth = $currentDate->copy()->endOfMonth()->min($to);

            $week = new Collection([
                'from' => $currentDate->copy()->startOfDay(), // range should start from the very firts second of the day
                'to' => $endOfCurrentMonth->copy()->endOfDay() // and end on the last second of the max day (23:59:59)
            ]);

            if ($currentIndex === 0) {
                $from->startOfMonth();
            }

            return $week;
        });  
    }

    public function getTheTotalNumberOfMonthsInRange()
    {
        (integer) $numberofMonths = 1;

        if ($this->getStartDate()->isSameYear($this->getEndDate()) && $this->getStartDate()->isSameMonth($this->getEndDate())) {
            return $numberofMonths;
        }

        $maxNumberofIterations = 1000;

        do {
            $numberofMonths++;
        } while (
            ($numberofMonths < $maxNumberofIterations) 
                && 
            (
                (
                    $this->getStartDate()
                      ->copy()
                      ->addMonthsNoOverflow($numberofMonths - 1)
                      ->isSameMonth($this->getEndDate()) === false
                ) 
                    ||
                (
                    $this->getStartDate()
                          ->copy()
                          ->addMonthsNoOverflow($numberofMonths - 1)
                          ->isSameMonth($this->getEndDate()) 
                          && 
                    !$this->getStartDate()
                          ->copy()
                          ->addMonthsNoOverflow($numberofMonths - 1)
                          ->isSameYear($this->getEndDate())
                )
            )
        );


        return $numberofMonths;
    }
    

    protected function hasForcedAUnit()
    {
        return in_array($this->forcedUnit, static::$validUnits);   
    }
    
    protected function validateArgments($forcedUnit)
    {
        if ($this->range->get('from')->greaterThan($this->range->get('to'))) {
            throw new InvalidArgumentException('Start date must be the same or lower than end date.');
        }

        if ($forcedUnit && !$this->hasForcedAUnit()) {
            throw new InvalidArgumentException("Invalid PeriodOfTime unit: {$forcedUnit}");
        } 
    }
    
}